//
// Programmer:    Craig Stuart Sapp <craig@ccrma.stanford.edu>
// Creation Date: 22 Nov 1998
// Last Modified: Sun Jan 10 00:05:07 PST 1999
// Filename:      cinmidi.cpp
// Syntax:        C++; sigControl
// $Smake:        g++ -DLINUX -Iinclude/sig %f %
//                -o %b -lpthread -Llib -lsig && strip %b
//

#include "improv.h"
#include <ctype.h>
#include <string.h>


// abbreviated general MIDI instrument names for the patch change
// interpretation display.  General MIDI instruments are standardized
// by the numbers 0-127.  If your instruments are not General MIDI, or
// you are using a instrument bank other than bank 0, then the names
// do not have any meaning.  These names are modified from the source:
//   ftp://ftp.linpeople.org/pub/People/nathan/playmidi-2.4.tar.gz
//
char *GMinstrument[128] = {
   "acpiano",   "britepno",  "synpiano",  "honkytonk", "epiano1",   "epiano2",
   "hrpschrd",  "clavinet",  "celeste",   "glocken",   "musicbox",  "vibes",
   "marimba",   "xylophon",  "tubebell",  "santur",    "homeorg",   "percorg",
   "rockorg",   "churchorg", "reedorg",   "accordn",   "harmonica", "concrtna",
   "nyguitar",  "acguitar",  "jazzgtr",   "cleangtr",  "mutegtr",   "odguitar",
   "distgtr",   "gtrharm",   "acbass",    "fngrbass",  "pickbass",  "fretless",
   "slapbas1",  "slapbas2",  "synbass1",  "synbass2",  "violin",    "viola",
   "cello",     "contraba",  "marcato",   "pizzcato",  "harp",      "timpani",
   "marcato",   "slowstr",   "synstr1",   "synstr2",   "choir",     "doo",
   "voices",    "orchhit",   "trumpet",   "trombone",  "tuba",      "mutetrum",
   "frenchorn", "hitbrass",  "synbras1",  "synbras2",  "sprnosax",  "altosax",
   "tenorsax",  "barisax",   "oboe",      "englhorn",  "bassoon",   "clarinet",
   "piccolo",   "flute",     "recorder",  "woodflut",  "bottle",    "shakazul",
   "whistle",   "ocarina",   "sqrwave",   "sawwave",   "calliope",  "chiflead",
   "charang",   "voxlead",   "lead5th",   "basslead",  "fantasia",  "warmpad",
   "polysyn",   "ghostie",   "bowglass",  "metalpad",  "halopad",   "sweeper",
   "aurora",    "soundtrk",  "crystal",   "atmosphr",  "freshair",  "unicorn",
   "sweeper",   "startrak",  "sitar",     "banjo",     "shamisen",  "koto",
   "kalimba",   "bagpipes",  "fiddle",    "shannai",   "carillon",  "agogo",
   "steeldrum", "woodblock", "taiko",     "toms",      "syntom",    "revcymb",
   "fx-fret",   "fx-blow",   "seashore",  "jungle",    "telephone", "helicptr",
   "applause",  "ringwhsl"
};


// global variables for command-line options:
Options   options;           // for command-line processing
int       absoluteQ = 0;     // for -a command-line option (time display)
int       interpretQ = 1;    // for -i command-line option (interpret input)
int       bytesQ = 1;        // no longer used, set to 1
int       style = 0;         // for -d, -x, and -2 command-line options
int       secondsQ = 0;      // for -s command-line option
int       keyboardQ = 1;     // for -k command-line option
int       fileQ = 0;         // for -o command-line option
int       suppressOffQ = 0;  // for -n command-line option
int       filter[8] = {0};   // for -f command-line option
int       cfilter[16] = {0}; // for -c command-line option
fstream   outputfile;        // for -o command-line option



// function declarations
void      checkOptions      (Options& opts);
void      displayMessage    (ostream& out, MidiMessage message, int style);
void      displayHeader     (ostream& out);
void      example           (void);
char*     getGmInst         (int number);
char*     getKey            (int keynum);
void      interpret         (ostream& out, MidiMessage message);
float     makePitchBend     (int lsb, int msb);
void      printbyte         (ostream& out, int value, int location, int style);
void      usage             (const char* command);


///////////////////////////////////////////////////////////////////////////

int main(int argc, char** argv) {
   options.setOptions(argc, argv);
   checkOptions(options);

   MidiInput messages(options.getInteger("port"));

   displayHeader(cout);
   if (fileQ) {
      displayHeader(outputfile);
   }

   KeyboardInput keyboard;     // for typing comments into output file
   char keych;                 // character from keyboard
   MidiMessage message;
   int lastTime = -1;

   while (1) {
      while (messages.getCount() > 0) {
         message = messages.extract();

         // filter any specified message types
         if (suppressOffQ && ((message.p0() & 0xf0) == 0x90) &&
               (message.p2() == 0)) {      
            continue;
         } else if (filter[(message.p0() >> 4) - 8]) {
            continue;
         } else if (cfilter[message.p0() & 0x0f]) {
            continue;
         }

         // adjust message time to delta time if necessary
         if (!absoluteQ) {
            if (lastTime == -1) {
               lastTime = message.time;
               message.time = 0;
            } else {
               int temp = message.time;
               message.time = message.time - lastTime;
               lastTime = temp;
            }
         }

         displayMessage(cout, message, style);
         if (fileQ) {
            displayMessage(outputfile, message, style);
         }
      }

      if (keyboardQ && keyboard.hit()) {
         keych = keyboard.getch();
         switch (keych) {
            case 27:   // the escape key 
               exit(0);      
               break;
            case 13:  // the enter key (this case needed for Microsoft)
               cout << endl;
               if (fileQ) {
                  outputfile << endl;
               }
               break;
            default:  
               cout << keych << flush;
               if (fileQ) {
                  outputfile << keych;
               }
         }
      }

      millisleep(1);   // sleep for 1 millisec for multi-tasking courtesy

   }

   return 0;
}


///////////////////////////////////////////////////////////////////////////


//////////////////////////////
//
// checkOptions -- process the command-line options
//

void checkOptions(Options& opts) {  // options are:
   opts.define("a|absolute=b");     //    boolean for absoute time display
   opts.define("i|interpret=b");    //    boolean for interpreting messages
   opts.define("p|port=i:0");       //    MIDI input port selection
   opts.define("l|list=b");         //    boolean for displaying possible inputs
   opts.define("d|dec|decimal=b");  //    boolean for printing bytes as decimal
   opts.define("x|hex|hexadecimal=b"); // boolean for printing bytes as hex
   opts.define("2|bin|binary=b");   //    boolean for printing bytes as binary
   opts.define("o|output=s");       //    for duplicating stdout to file
   opts.define("n|nooffs=b");       //    boolean for suppressing note-offs
   opts.define("f|filter=s:");      //    for filtering out various commands
   opts.define("c|chan-filter=s:"); //    for filtering out various channels
   opts.define("u|user=s");         //    for displaying in header info
   opts.define("s|seconds=b");      //    for displaying time info in seconds
   opts.define("k|keyboard=b");     //    for handling computer keyboard
   opts.define("author=b");
   opts.define("version=b");
   opts.define("example=b");
   opts.define("h|help=b");
   opts.process();

   if (opts.getBoolean("author")) {
      cout << "Written by Craig Stuart Sapp, "
              "craig@ccrma.stanford.edu, November 1998" << endl;
      exit(0);
   } else if (opts.getBoolean("version")) {
      cout << "cinmidi, version 1.0 (25 Jan 1999)\n"
              "compiled: " << __DATE__ << endl;
      exit(0);
   } else if (opts.getBoolean("help")) {
      usage(opts.getCommand());
      exit(0);
   } else if (opts.getBoolean("example")) {
      example();
      exit(0);
   }


   if (opts.getBoolean("list")) {
      // display the possible MIDI inputs to use with -p option
      cout << "\nMidi Input ports: " << MidiInput::getNumPorts() << endl;
      for (int i=0; i<MidiInput::getNumPorts(); i++) {
         cout << "\t" << i << ": " << MidiInput::getName(i) << '\n';
      }
      cout << endl;
      exit(0);
   }


   if (opts.getBoolean("absolute")) {
      absoluteQ = 1;
   }
   if (opts.getBoolean("interpret")) {
      interpretQ = 0;
   }
   if (opts.getBoolean("decimal")) {
      style = 10;
   } else if (opts.getBoolean("hexadecimal")) {
      style = 16;
   } else if (opts.getBoolean("binary")) {
      style = 2;
   } else {
      style = 0;
   }
   if (style != 0) {
      bytesQ = 1;
   }

   if (opts.getBoolean("nooffs")) {
      suppressOffQ = 1;
   }
   if (opts.getBoolean("seconds")) {
      secondsQ = 1;
   }
   if (opts.getBoolean("keyboard")) {
      keyboardQ = 0;
   }
   
   const char *filterString = opts.getString("filter");
   char current;
   for (int i=0; i<(int)strlen(filterString); i++) {
      current = filterString[i];
      switch (tolower(current)) {
         case '8':   filter[0] = 1;   break;
         case '9':   filter[1] = 1;   break;
         default:
            if (!isdigit(current) && isxdigit(current)) {
               if (current-'a'+2 >= 8 || current-'a'+2 < 0) {
                  cerr << "Error in filter string" << endl;
                  exit(1);
               } else {
                  filter[current - 'a' + 2] = 1;
               }
            }
      }
   }
   
   const char *cfilterString = opts.getString("chan-filter");
   char stringc [2] = {0};
   int cindex;
   for (int j=0; j<(int)strlen(cfilterString); j++) {
      if (isxdigit(cfilterString[j])) {
         stringc[0] = cfilterString[j];
         cindex = strtol(stringc, NULL, 16);
         if (cindex < 0 || cindex > 15) {
            cout << "Error in channel string: " << cfilterString << endl;
            exit(1);
         } else {
            cfilter[cindex] = 1;
         }
      }
         
   }

   if (opts.getBoolean("output")) {
      fileQ = 1;
      outputfile.open(opts.getString("output"), ios::out);
      if (!outputfile.is_open()) {
         cout << "Error: cannot open file: "<< opts.getString("output") << endl;
         exit(1);
      }
   }


   if (opts.getArgCount() != 0) {
      usage(opts.getCommand());
      exit(1);
   }
}

  
//////////////////////////////
//
// displayHeader -- display data formatting information
//

void displayHeader(ostream& out) {
   out << "; \n";
   if (options.getBoolean("user")) {
      out << "; Recorded by:    ";
      out << options.getString("user") << endl;
   }
   out << "; Style:          ";
   switch (style) {
      case 10:   out << "decimal\n";       break;
      case 16:   out << "hexadecimal\n";   break;
      case  2:   out << "binary\n";        break;
      default:   out << "default\n"; 
   }
   out << "; Timing:         ";
   if (absoluteQ) {
      out << "absolute";
      if (secondsQ) {
         out << " seconds\n";
      } else {
         out << " milliseconds\n";
      }
      out << "; Message format: received-time, MIDI command-byte, "
          "MIDI parameter-byte(s)\n";
   } else {
      out << "delta";
      if (secondsQ) {
         out << " seconds\n";
      } else {
         out << " milliseconds\n";
      }
      out << "; Message format: delta-time, MIDI command-byte, "
          "MIDI parameter-byte(s)\n";
   }
   out << "; Version:        ascmidi 1.0\n";
   out << "; Command-line:   " << options.getCommandLine() << endl;
   out << "; Input Port:     " << options.getInteger("port")
       << ": " << MidiInput::getName(options.getInteger("port"))
       << endl;
   out << "; \n";
   out << endl;
}



//////////////////////////////
//
// displayMessage -- displays the MIDI input message in the style
//	specified on the command-line
//

void displayMessage(ostream& out, MidiMessage message, int style) {
   if (secondsQ) {
      out << dec << message.time / 1000.0 << "\t";
   } else {
      out << dec << message.time << "\t";
   }
   
   if (bytesQ) {
      printbyte(out, (int)message[0], 0, style);
      int parameterCount = message.getParameterCount();
      for (int i=0; i<parameterCount; i++) {
         out << " ";
         printbyte(out, (int)message[i+1], i+1, style);
      }
   }

   if (bytesQ && interpretQ && (style == 2)) {
      if (message.getParameterCount() == 1) {
         out << '\t';
      }
      out << "\t ; ";
   } else if (bytesQ && interpretQ) {
      if (message.getParameterCount() == 1 && (style != 0)) {
         out << '\t';
      }
      out << "\t\t; ";
   }

   if (interpretQ) {
      interpret(out, message);
   }

   out << endl;
}



//////////////////////////////
//
// example -- shows various command-line option calls to program.
//

void example(void) {
   cout << 
   "# display MIDI input data in binary form:\n"
   "cinmidi -2\n"
   "\n"
   "# don't display note-off messages:\n"
   "cinmidi -n\n"
   "\n"
   "# don't display pitchbend data:\n"
   "cinmidi -fe\n"
   "\n"
   "# display only pitchbend data:\n"
   "cinmidi -f 89abcdf\n"
   "\n"
   "# display only note-on messages:\n" 
   "cinmidi -n -f 8abcdef\n"
   "\n"
   << endl;
}



//////////////////////////////
//
// getGmInst -- the General MIDI instrument name.
//

char* getGmInst(int number) {
   return GMinstrument[number % 0x80];
}



//////////////////////////////
//
// getKey -- return the key name of the specified MIDI key number.
//

char* getKey(int keynum) {
   static char name[16] = {0};
   static char temp[16] = {0};
   int octave = keynum / 12 - 1;

   switch (keynum % 12) {
      case  0:   strcpy(name, "C");    break;
      case  1:   strcpy(name, "C#");   break;
      case  2:   strcpy(name, "D");    break;
      case  3:   strcpy(name, "Eb");   break;
      case  4:   strcpy(name, "E");    break;
      case  5:   strcpy(name, "F");    break;
      case  6:   strcpy(name, "F#");   break;
      case  7:   strcpy(name, "G");    break;
      case  8:   strcpy(name, "Ab");   break;
      case  9:   strcpy(name, "A");    break;
      case 10:   strcpy(name, "Bb");   break;
      case 11:   strcpy(name, "B");    break;
      default:   strcpy(name, "");
   }

   sprintf(temp, "%d", octave);
   strcat(name, temp);

   return name;
}



//////////////////////////////
//
// interpret -- convert a MIDI message to English.
//

void interpret(ostream& out, MidiMessage message) {
   out << dec;      // set output to decimal notation
 
   switch (message.p0() & 0xf0) {
      case 0x80:                                           // NOTE OFF
         out << "NOTEOFF chan: " << (int)(message.p0() & 0x0f) + 1
             << " key:" << getKey(message.p1()) << " vel: "
             << (int)message.p2();
         break;
      case 0x90:                                           // NOTE ON
         out << "NOTE chan:" << (int)(message.p0() & 0x0f) + 1
             << " key:" << getKey(message.p1())
             << " vel:" << (int)message.p2();
         if (message.p2() == 0) {
            out << " OFF";
         }
         break;
      case 0xa0:                                           // AFTERTOUCH
         out << "AFTERTOUCH chan:" << (int)(message.p0() & 0x0f) + 1
             << " key:" << getKey(message.p1())
             << " val:" << (int)message.p2();
         break;
      case 0xb0:
         out << "CONTROL chan:" << (int)(message.p0() & 0x0f) + 1
             << " #:" << (int)message.p1()
             << " val:" << (int)message.p2();
         switch (message.p1()) {
            case  7:  out << " (volume)";   break;
            case 64:  out << " (sustain)";  break;
         }
         break;
      case 0xc0:                                           // PATCH CHANGE
         out << "PATCH chan:" << (int)(message.p0() & 0x0f) + 1
             << " inst:" << (int)message.p1()
             << " (" << getGmInst(message.p1()) << ")";
         break;
      case 0xd0:                                           // CHANNEL PRESSURE
         out << "PRESSURE chan:" << (int)(message.p0() & 0x0f) + 1
             << " val:" << (int)message.p1();
         break;
      case 0xe0:                                           // PITCH BEND
         out << "PITCHBEND chan:" << (int)(message.p0() & 0x0f) + 1
             << " val:" << makePitchBend(message.p1(), message.p2());
         break;
      case 0xf0:
         out << "SYSTEM MESSAGE";
         break;
   }
}
         


//////////////////////////////
//
// makePitchBend -- create a number in range -1 to 1.
//

float makePitchBend(int lsb, int msb) {
   int value = (msb << 7) | lsb;
   float output = (float)((2.0 * value / 0x3fff) - 1.0);
   if ((output < 0.0001) && (output > -0.0001)) {
      return 0.0;
   } else {
      output = (float)(((int)(output * 1000.0))/1000.0);
      return output;
   }
}



//////////////////////////////
//
// printbyte -- display a byte according to specified style and location.
//

void printbyte(ostream& out, int value, int location, int style) {
   int i;
   switch (style) {
      case 10: 
         if (value < 100) {
            out << ' ';
         } 
         if (value < 10) {
            out << ' ';
         }
         out << dec << value;
         break;
      case 16:
         if (value < 0x10) {
            out << ' ';
         }
         out << hex << value;
         break;
      case  2:
         for (i=0; i<8; i++) {
            if (value & (1 << (7-i))) {
               out << '1';
            } else {
               out << '0';
            }
         }
         break;
      default:
         if (location == 0) {
            out << "0x";
            printbyte(out, value, 0, 16);
         } else {
            printbyte(out, value, location, 10);
         }
   }
}


   
//////////////////////////////
//
// usage -- how to run this program from the command-line.
//

void usage(const char* command) {
   cout << 
   "\n"
   "Display/Record MIDI input data.\n"
   "\n"
   "Usage: " << command << " [-a][-i][-x|-d|-2] [-o output][-p port]\n"
   "\n"
   "Options:\n"
   "   -a = display MIDI timing in absolute time instead of delta time\n"
   "   -i = don't display interpretation of MIDI message as comments\n"
   "   -d = display MIDI bytes in decimal form\n"
   "   -x = display MIDI bytes in hexadecimal form\n"
   "   -2 = display MIDI bytes in binary form\n"
   "   -n = do not display note-off messages\n"
   "   -f string = filter out specified MIDI commands (e.g. \"bc\" will\n"
   "        filter out control change and patch change messages)\n"
   "   -c string = filter out specified MIDI channels, offset zero\n"
   "        (e.g., \"09AF\", filters channels 1, 10, 11, and 16)\n"
   "   -p port = use the specified MIDI input port\n"
   "   -l = list MIDI input ports and then exit\n"
   "   -o filename = send display to file as well as to screen\n"
   "   -u string = specify a user name for the header information\n"
   "   -s = display time values in seconds\n"
   "   -k = disable keyboard input\n"
   "   --options = list all options, default values, and aliases\n"
   "\n"
   << endl;
}




