//
// Programmer:    Craig Stuart Sapp <craig@ccrma.stanford.edu> 
// Creation Date: 18 December 1997
// Last Modified: Mon Jan 26 23:54:36 GMT-0800 1998
// Filename:      ...sig/code/control/MidiOutput/MidiOutput.cpp
// Syntax:        C++
// $Smake:        cc -g -c %b.cc -I../../../include && rm -f %b.o
// 
// Description: The MIDI output interface for MIDI synthesizers/equipment
//


#include "MidiOutput.h"
#include <iostream.h>
#include <iomanip.h>

#define RECORD_ASCII     (0)
#define RECORD_BINARY    (1)
#define RECORD_MIDI_FILE (2)


// declaration of static variables
SigTimer MidiOutput::timer;


//////////////////////////////
//
// MidiOutput::MidiOutput
//


MidiOutput::MidiOutput(void) : MidiOutPort() {
   outputRecordQ = 0;
}


MidiOutput::MidiOutput(int aPort, int autoOpen) : MidiOutPort(aPort, autoOpen) {
   // do nothing
}



//////////////////////////////
//
// MidiOutput::~MidiOutput
//

MidiOutput::~MidiOutput() {
   // do nothing
}



//////////////////////////////
//
// MidiOutput::cont -- send a controller command MIDI message.
//
//    channel = the Midi channel ofset from 0 [0..15]
//    controller = the continuous controller number [0..127]
//    data = the value of the specified controller [0..127]
//

int MidiOutput::cont(int channel, int controller, int data) {
   return send(0xb0 | (channel & 0x0f), controller, data);
}



//////////////////////////////
//
// MidiOutput::off -- sends a Note Off MIDI message (0x80).
//
//    channel = MIDI channel to send note on. range is [0..15]
//    keynum = MIDI key number to play (middle C = 60, C# = 61, etc.) [0..127]
//    velocity = release velocity of the note, 127 = quickest possible
//
// Note: The more common method of turning off a note is to use the
//	play() function (midi command 0x90) but send an attack velocity of 0.
//

int MidiOutput::off(int channel, int keynum, int releaseVelocity) {
   return send(0x80 | (channel & 0x0f), keynum, releaseVelocity);
}



//////////////////////////////
//
// MidiOutput::pc -- send a patch change MIDI message. changes the timbre
//    on the specified channel.  
//
//    channel = MIDI channel to which to send the patch change [0..15]
//    timbre = the voice to select on the specified channel [0..127]
//

int MidiOutput::pc(int channel, int timbre) {
   return send(0xc0 | (channel & 0x0f), timbre, 0);
}



//////////////////////////////
//
// MidiOutput::play -- sends a Note On/Off MIDI message.
//
//    channel = MIDI channel to send note on. range is [0..15]
//    keynum = MIDI key number to play (middle C = 60, C# = 61, etc.) [0..127]
//    velocity = attack velocity of the note, 0 = 0ff, 127 = loudest possible
//

int MidiOutput::play(int channel, int keynum, int velocity) {
   return send(0x90 | (channel & 0x0f), keynum, velocity);
}



//////////////////////////////
//
// MidiOutput::pw -- Pitch Wheel: send a MIDI pitch bend.
//      Parameters are:
//         1. channel   -- MIDI channel offset from 0.
//         2. mostByte  -- most significant 7 bits (coarse tuning)
//         3. leastByte -- least significant 7 bits (fine tuning)
//

int MidiOutput::pw(int channel, int mostByte, int leastByte) {
   return send(0xe0 | (channel & 0x0f), leastByte, mostByte);
}



//////////////////////////////
//
// MidiOutput::pw -- Pitch Wheel: 14 bit number given as input
//      range for 14 bit number is 0 to 16383.
//

int MidiOutput::pw(int channel, int tuningData) {
   uchar greaterBits = (uchar)((tuningData >> 7) & 0x7f);
   uchar lesserBits = (uchar)(tuningData & 0x7f);
   return pw(channel, greaterBits, lesserBits);
}



//////////////////////////////
//
// MidiOutput::pw -- Pitch Wheel: range between -1 to 1 given as input.
//      Range is then converted to a 14 bit number.
//      +1 = highest value of pitch wheel
//       0 = rest position of pitch wheel
//      -1 = lowest value of pitch wheel
//

int MidiOutput::pw(int channel, double tuningData) {
   if (tuningData < -1.0 || tuningData > 1.0) {
      cerr << "Error: pitch wheel data is out of range: " << tuningData << endl;
      exit(1);
   }

   int output = (int)((tuningData+1.0)/2.0*16383 + 0.5);
   return pw(channel, output);
}



//////////////////////////////
//
// MidiOutput::recordStart
//

void MidiOutput::recordStart(char *filename, int format) {
   if (outputRecordQ) {  // already recording, so close old file
      recordStop();
   }

   outputRecordFile.open(filename, ios::out);
   if (!outputRecordFile) {   // open file failed
      cerr << "Error: cannot open file " << filename << endl;
      outputRecordQ = 0;
   } else {
      outputRecordQ = 1;
   }

   if (outputRecordQ) {
      switch (format) {
         case RECORD_ASCII:   // ascii
            outputRecordType = RECORD_ASCII;
            outputRecordFile <<"; delta time/MIDI output at delta time" << endl;
            break;
         case RECORD_BINARY:   // binary
            outputRecordType = RECORD_BINARY;
            // record the magic number for binary format
            outputRecordFile << (uchar)0xf8 << (uchar)0xf8
                       << (uchar)0xf8 << (uchar)0xf8;
            break;
         case RECORD_MIDI_FILE: // standard MIDI file, type 0
         default:
            outputRecordType = RECORD_MIDI_FILE;
            // header stuff to be written here
            break;
      }
   }

   lastFlushTime = timer.getTime();
}



//////////////////////////////
//
// MidiOutput::recordStop
//

void MidiOutput::recordStop(void) {
   if (outputRecordQ) {
      outputRecordQ = 0;
      outputRecordFile.close();
   }
}



//////////////////////////////
//
// MidiOutput::reset -- sends the MIDI command 0xFF which
//      should force the MIDI devices on the other side of the
//      MIDI cable into their power-on reset condition, clear running
//      status, turn off any sounding notes, set Local Control on, and
//      otherwise clean up the state of things.
//

void MidiOutput::reset(void) {
   send(0xff, 0, 0);
}



//////////////////////////////
//
// MidiOutput::send -- send a byte to the MIDI port but record it
//	first.
//

int MidiOutput::send(int command, int p1, int p2) {
   if (outputRecordQ) {
      switch (outputRecordType) {
         case 0:   // ascii
            writeOutputAscii(command, p1, p2);
            break;
         case 1:   // binary
            writeOutputBinary(command, p1, p2);
            break;
         case 2:   // standard MIDI file type 0
            writeOutputMidifile(command, p1, p2);
            break;
      }
      lastFlushTime = timer.getTime();  // only keep track if recording
   }
   return rawsend(command, p1, p2);
}


int MidiOutput::send(int command, int p1) {
   if (outputRecordQ) {
      switch (outputRecordType) {
         case 0:   // ascii
            writeOutputAscii(command, p1, -1);
            break;
         case 1:   // binary
            writeOutputBinary(command, p1, -1);
            break;
         case 2:   // standard MIDI file type 0
            writeOutputMidifile(command, p1, -1);
            break;
      }
      lastFlushTime = timer.getTime();  // only keep track if recording
   }
   return rawsend(command, p1);
}


int MidiOutput::send(int command) {
   if (outputRecordQ) {
      switch (outputRecordType) {
         case 0:   // ascii
            writeOutputAscii(command, -1, -1);
            break;
         case 1:   // binary
            writeOutputBinary(command, -1, -1);
            break;
         case 2:   // standard MIDI file type 0
            writeOutputMidifile(command, -1, -1);
            break;
      }
      lastFlushTime = timer.getTime();  // only keep track if recording
   }
   return rawsend(command);
}



//////////////////////////////
//
// MidiOutput::silence -- send a note off to all notes on all channels.
//    default value: aChannel = -1
//

void MidiOutput::silence(int aChannel) {
   int channel, keyno;
   if (aChannel == -1) {
      for (channel=0; channel<16; channel++) {
         for (keyno=0; keyno<128; keyno++) {
            play(keyno, 0, channel);
         }
      }
   } else {
      for (keyno=0; keyno<128; keyno++) {
         play(keyno, 0, aChannel);
      }
   }
}



//////////////////////////////
//
// MidiOutput::sustain -- set the MIDI sustain continuous controller on or off.
//	Equivalent to the command cont(channel, 0x40, status).
//

void MidiOutput::sustain(int channel, int status) {
   if (status) {  // turn on sustain
      cont(channel, 0x40, 127);
   } else {       // turn off sustain
      cont(channel, 0x40, 0);
   }
}



/*
//////////////////////////////
//
// MidiOutput::sysex -- sends a system exclusive MIDI message.
//    data = array of data bytes.  each byte must be in the range [0..127]
//    length = number of data bytes in the array.
//

int MidiOutput::sysex(char* data, int length) {
   MIDIHDR header;
   header.lpData = (char*)data;
   header.dwBufferLength = length;
   header.dwFlags = 0;
  
   int flag = midiOutPrepareHeader(midiOutDevice, &header, sizeof(MIDIHDR));
   if (flag == MMSYSERR_NOERROR) {
      midiOutLongMsg(midiOutDevice, &header, sizeof(MIDIHDR));
      midiOutUnprepareHeader(midiOutDevice, &header, sizeof(MIDIHDR));
   }

   return flag;
}

*/




///////////////////////////////////////////////////////////////////////////
//
// private functions
//


//////////////////////////////
//
// MidiOutput::writeOutputAscii
//

void MidiOutput::writeOutputAscii(int command, int p1, int p2) {
   outputRecordFile << dec;
   outputRecordFile.width(6);
   outputRecordFile << (timer.getTime()-lastFlushTime) <<'\t';
   outputRecordFile << "0x" << hex;
   outputRecordFile.width(2);
   outputRecordFile << command << ' ';
   outputRecordFile << dec; 
   outputRecordFile.width(3);
   outputRecordFile << p1 << ' ';
   outputRecordFile << dec; 
   outputRecordFile.width(3);
   outputRecordFile<< p2;
   outputRecordFile << endl;
}



//////////////////////////////
//
// MidiOutput::writeOutputBinary
//

void MidiOutput::writeOutputBinary(int command, int p1, int p2) {
   // don't store 0xf8 command since it will be used to mark the end of the 
   if (command == 0xf8) return;

   // write the delta time (four bytes)
   outputRecordFile.writeBigEndian((ulong)(timer.getTime() - lastFlushTime));

   // write midi data 
   // don't store 0xf8 command since it will be used to mark the end of the 
   // delta time data.
   outputRecordFile << (uchar)p1;
   outputRecordFile << (uchar)p2;
   outputRecordFile << (uchar)0xf8;
}



//////////////////////////////
//
// MidiOutput::writeOutputMidifile
//

void MidiOutput::writeOutputMidifile(int command, int p1, int p2) {
   // not yet implemented
}




