//
// Programmer:    Craig Stuart Sapp <craig@ccrma.stanford.edu>
// Creation Date: 21 December 1997
// Last Modified: Wed Feb 11 22:43:15 GMT-0800 1998
// Last Modified: Wed Sep 30 13:22:51 PDT 1998
// Filename:      ...sig/code/control/RadioDrum/RadioDrum.cpp
// Syntax:        C++
// $Smake:        cc -Wall -g -c %b.cc -I../../../include && rm -f %b.o
//
// Description: A structure for handling the RadioDrum variables.
//


#include <ctype.h>
#include <iomanip.h>

#ifdef LINUX
   #include <unistd.h>
#endif
#ifdef SUN
   #include <unistd.h>
#endif
#ifdef IRIX
   #include <unistd.h>
#endif


///////////////////////////////////////////////////////////////////////////


#include <fstream.h>
#include "RadioDrum.h"


//////////////////////////////
// 
// RadioDrum::RadioDrum
//

RadioDrum::RadioDrum(void) : MidiIO() {
   setStateSize(DEFAULT_STATE_SIZE);
   for (int i=0; i<16; i++) {
      completeBufQ[i] = 0;
   }
   for (int j=0; j<22; j++) {
      calib[j] = 0;
   }
   reportingQ = 0;
   recordStateQ = 0;
   storedReportStatus = 0;
   errorQ = 0;
   writingCalib = 0;
   calibFilename = NULL;
   calibFilename = new char[100];
   strcpy(calibFilename, "calib.def");
   
   stick1trig = RadioBatonEmptyBehavior;
   stick2trig = RadioBatonEmptyBehavior;
   stick1position = RadioBatonEmptyBehavior;
   stick2position = RadioBatonEmptyBehavior;
   b14plustrig = RadioBatonEmptyBehavior;
   b15plustrig = RadioBatonEmptyBehavior;
   b14minusuptrig = RadioBatonEmptyBehavior;
   b14minusdowntrig = RadioBatonEmptyBehavior;
   b15minusuptrig = RadioBatonEmptyBehavior;
   b15minusdowntrig = RadioBatonEmptyBehavior;
}


RadioDrum::RadioDrum(int outputDevice, int inputDevice) :
      MidiIO(outputDevice, inputDevice) {
   setStateSize(DEFAULT_STATE_SIZE);
   for (int i=0; i< 16; i++) {
      completeBufQ[i] = 0;
   }
   for (int j=0; j<22; j++) {
      calib[j] = 0;
   }
   reportingQ = 0;
   recordStateQ = 0;
   storedReportStatus = 0;
   errorQ = 0;
   writingCalib = 0;
   calibFilename = NULL;
   calibFilename = new char[100];
   strcpy(calibFilename, "calib.def");
   
   stick1trig = RadioBatonEmptyBehavior;
   stick2trig = RadioBatonEmptyBehavior;
   b14plustrig = RadioBatonEmptyBehavior;
   b15plustrig = RadioBatonEmptyBehavior;
   b14minusuptrig = RadioBatonEmptyBehavior;
   b14minusdowntrig = RadioBatonEmptyBehavior;
   b15minusuptrig = RadioBatonEmptyBehavior;
   b15minusdowntrig = RadioBatonEmptyBehavior;
}



//////////////////////////////
//
// RadioDrum::~RadioDrum
//

RadioDrum::~RadioDrum() {
   if (calibFilename != NULL) {
      delete [] calibFilename;
      calibFilename = NULL;
   }
   if (recordStateQ == 1) {
      recordStateStop();
      recordStateQ = 0;
   }
}



//////////////////////////////
//
// RadioDrum::centerStick1
//

void RadioDrum::centerStick1(void) {
   sendMessage(SET_CENTER_STICK1);
}



//////////////////////////////
//
// RadioDrum::centerStick2
//

void RadioDrum::centerStick2(void) {
   sendMessage(SET_CENTER_STICK2);
}



//////////////////////////////
//
// RadioDrum::checkCalibration -- must have monotonically increasing
//     values in calib array locations 10-20, and values must be less 
//     than 2048.
//

int RadioDrum::checkCalibration(void) {
   int value = calib[10];
   for (int i = 11; i<21; i++) {
      if (calib[i] <= value || calib[i] >= 2048) {
         return 0;
      }
      value = calib[i];
   }
   return 1;
}



//////////////////////////////
//
// RadioDrum::decreaseXYSensitivity
//
 
void RadioDrum::decreaseXYSensitivity(void) {
   sendMessage(DEC_XY_SENSITIVITY);
}



//////////////////////////////
//
// RadioDrum::decreaseZSensitivity
//

void RadioDrum::decreaseZSensitivity(void) {
   sendMessage(DEC_Z_SENSITIVITY);
}



//////////////////////////////
//
// RadioDrum::getBuf -- returns the currently
//    stored value for the given buffer.  Returns
//    -1 if the radio baton has not completed sending the 
//    value data, which is a 12-bit unsigned value
//    sent in 4 nibbles (most significant nibble is always 0).
//    Will also return -1 if an invalid array location is
//    accessed.
//

int RadioDrum::getBuf(int index) {
   if (index < 0 || index > 15) {
      cerr << "Invalid array index for buf: " << index << endl;
      return -1;
   }

   if (completeBufQ[index] == 0) {
      return -1;
   } else {
      return buf[index];
   }
}



//////////////////////////////
//
// RadioDrum::getError -- return the state of the 
//    MIDI input error printing value
//

int RadioDrum::getError(void) const {
   return errorQ;
}



//////////////////////////////
//
// RadioDrum::getPositionReporting
//

int RadioDrum::getPositionReporting(void) const {
   return reportingQ;
}



//////////////////////////////
//
// RadioDrum::getReportStatus
//

int RadioDrum::getReportStatus(void) const {
   return reportingQ;
}



//////////////////////////////
//
// RadioDrum::getWhack1Time -- returns the 
//    most recent trigger time of stick 1.
//

long RadioDrum::getWhack1Time(void) const {
   return t1t;
}



//////////////////////////////
//
// RadioDrum::getWhack2Time -- returns the more recent trigger 
//     time of stick 2
//

long RadioDrum::getWhack2Time(void) const {
   return t2t;
}



//////////////////////////////
//
// RadioDrum::increaseXYSensitivity(void) {
//
 
void RadioDrum::increaseXYSensitivity(void) {
   sendMessage(INC_XY_SENSITIVITY);
}



//////////////////////////////
//
// RadioDrum::increaseZSensitivity
//

void RadioDrum::increaseZSensitivity(void) {
   sendMessage(INC_Z_SENSITIVITY);
}



//////////////////////////////
//
// RadioDrum::positionReportingOff -- returns the previously
//     set state of position reporting.
//

int RadioDrum::positionReportingOff(void) {
   int oldState = reportingQ;
   stoprunningstatus();
   rawsend(BAT_MIDI_COMMAND, REPORT_POS, POS_REPORT_OFF);
   reportingQ = 0;
   return oldState;
}



//////////////////////////////
// 
// RadioDrum::positionReportingOn -- returns the previously
//     set state of position reporting.
//

int RadioDrum::positionReportingOn(void) { 
   int oldState = reportingQ;
   stoprunningstatus();
   rawsend(BAT_MIDI_COMMAND, REPORT_POS, POS_REPORT_ON);
   reportingQ = 1;
   return oldState;
}



//////////////////////////////
//
// RadioDrum::processIncomingMessages
// 

void RadioDrum::processIncomingMessages(void) {
   while (MidiInput::getCount() > 0) {
      interpretCommand(MidiInput::extract());
   }
}



//////////////////////////////
//
// RadioDrum::readCalibrationFile -- send calibration data from
//    radio baton to the specified file. If file couldn't be
//    found, return 0; otherwise, returns 1.
//

int RadioDrum::readCalibrationFile(const char* aFilename) {
   fstream input(aFilename, ios::in | ios::nocreate);
   if (!input) {
      return 0;
   }
   char tempString[256];
   
   input >> tempString;     // hit1
   input >> calib[0];
   input >> tempString;     // set1
   input >> calib[1];
   input >> tempString;     // hit2
   input >> calib[2];
   input >> tempString;     // set2
   input >> calib[3];
   input >> tempString;     // zero1x
   input >> calib[4];
   input >> tempString;     // zero1y
   input >> calib[5];
   input >> tempString;     // zero2x
   input >> calib[6];       
   input >> tempString;     // zero2y
   input >> calib[7];
   input >> tempString;     // range
   input >> calib[8];       
   input >> tempString;     // restpoint
   input >> calib[9];
   for (int i=10; i<21; i++) {
      input >> calib[i];
   }
   input >> tempString;      // zcalib
   input >> calib[21];
 
   // send calibration data to the radio drum only if calibration file is 
   // read completely.  Check by comparing last string which should be "zcalib".
   for (int j=0; j<6; j++) {
      if (tempString[j] == '\0') break;
      tempString[j] = (char)tolower(tempString[j]);
   }
   if (checkCalibration() == 0) {
      cerr << "Calibration file is corrupt; file " << aFilename
           << " was not sent." << endl;
      return 0;
   }
   if (strcmp("zcalib", tempString) == 0 ) {
      // now tell the radio drum what the calibration is:
      sendCalibration();      

      return 1;       // successful reading of calibration file.
   } else {
      return 0;       // unsuccessful reading of calibration file.
   }
}



//////////////////////////////
//
// RadioDrum::recordingQ -- returns true if currently recording,
//	false if not recording.
//

int RadioDrum::recordingQ(void) const {
   return recordStateQ;
}



//////////////////////////////
//
// RadioDrum::recordStateStart -- 
//

void RadioDrum::recordStateStart(const char* aFileName) {
   if (recordStateQ == 1) {
      recordStateStop();
   }

   recordOutput.open(aFileName, ios::out);
   if (!recordOutput) {   // open file failed
      cerr << "Error: cannot open file " << aFileName << endl;
      recordStateQ = 0;
   } else {
      recordStateQ = 1;
      timer.reset();
      recordTimeOffset = -1;
   }
}



//////////////////////////////
//
// RadioDrum::recordStateStop
//

void RadioDrum::recordStateStop(void) {
   if (recordStateQ == 0)  return;
   if (recordOutput.is_open()) {
      recordOutput.close();
      recordTimeOffset = -1;
      recordStateQ = 0;
   }
}



//////////////////////////////
//
// RadioDrum::requestCalibration -- Send a request for calibration data 
//    from the radio drum.  It will be received later from the 
//    radio drum and stored in the calib array.
//

void RadioDrum::requestCalibration(void) {
   stoprunningstatus(); 
   rawsend(BAT_MIDI_COMMAND, REPORT_CALIB, 0);
}



//////////////////////////////
//
// RadioDrum::requestBufData -- Gets all 16 values of buf data
//    from the radio drum.
//

void RadioDrum::requestBufData(void) {
   // send request for buffer data
   for (int i=0; i<16; i++) {
      stoprunningstatus();
      rawsend(BAT_MIDI_COMMAND, REPORT_BUF, i);
   }
   // the processIncomingMessages function will detect the results
   // of this buff data request and fill the buff array.
}



//////////////////////////////
//
// RadioDrum::sendMessage -- sends a message to the radio drum.
//    Messages are those related to the BAT_SEND set
//    found in baton.h .
//

void RadioDrum::sendMessage(int aMessage) {
   if (aMessage == TEST_BATON) {
      testBaton();
   } else {
      stoprunningstatus();
      rawsend(BAT_MIDI_COMMAND, BAT_SEND, aMessage);
   }
}



//////////////////////////////
//
// RadioDrum::setError -- if true, then will print out
//    unknown MIDI commands which seem to be bad.
//    In other words, if an 0xa0 command byte was 
//    received, and it doesn't match any baton messages,
//    then an error message will be print to cerr.
//    If false, then don't print any message for bad
//    commands.
//

void RadioDrum::setError(int errorStatus) {
   if (errorStatus) {
      errorQ = 1;
   } else {
      errorQ = 0;
   }
}



//////////////////////////////
//
// RadioDrum::setReportStatus
//

void RadioDrum::setReportStatus(int aStatus) {
   if (aStatus == 0) {
      reportingQ = 0;
      positionReportingOff();
   } else {
      reportingQ = 1;
      positionReportingOn();
   }
}


//////////////////////////////
//
// RadioDrum::setStateSize -- set the length of the
//    histories of each state variable in the object.
//

void RadioDrum::setStateSize(int aSize) {
   if (aSize <= 0) {
      cerr << "Error: state size too small: " << aSize << endl;
      exit(1);
   }

   t1pb.setSize(aSize);    // stick1 position time
   x1pb.setSize(aSize);    // stick1 x-axis position
   y1pb.setSize(aSize);    // stick1 y-axis position
   z1pb.setSize(aSize);    // stick1 z-axis position

   t2pb.setSize(aSize);    // stick2 position time
   x2pb.setSize(aSize);    // stick2 x-axis position
   y2pb.setSize(aSize);    // stick2 y-axis position
   z2pb.setSize(aSize);    // stick2 z-axis position   

   d1pb.setSize(aSize);    // dial1 position
   d2pb.setSize(aSize);    // dial2 position
   d3pb.setSize(aSize);    // dial3 position
   d4pb.setSize(aSize);    // dial4 position

   t1tb.setSize(aSize);    // stick1 trigger time
   x1tb.setSize(aSize);    // stick1 x-axis trigger pos.
   y1tb.setSize(aSize);    // stick1 y-axis trigger pos.
   w1tb.setSize(aSize);    // stick1 wack at trigger time

   t2tb.setSize(aSize);    // stick2 trigger time
   x2tb.setSize(aSize);    // stick2 x-axis trigger pos.
   y2tb.setSize(aSize);    // stick2 y-axis trigger pos.
   w2tb.setSize(aSize);    // stick2 whach at trigger time

   b14ptb.setSize(aSize);  // b14+ button trigger time buffer
   b15ptb.setSize(aSize);  // b15+ button trigger time buffer
   b14mdtb.setSize(aSize); // b14- pedal down trigger time buffer
   b14mutb.setSize(aSize); // b14- pedal up trigger time buffer
   b15mdtb.setSize(aSize); // b15- pedal down trigger time buffer
   b15mutb.setSize(aSize); // b15- pedal up trigger time buffer
}



//////////////////////////////
//
// RadioDrum::setStickLevels
//

void RadioDrum::setStickLevels(void) {
   sendMessage(SET_STICK_LEVELS);
}



//////////////////////////////
//
// RadioDrum::testBaton -- checks to see if radio drum is connected
//

int RadioDrum::testBaton(void) {
   int reportStatus = positionReportingOff();
   stoprunningstatus();
   rawsend(BAT_MIDI_COMMAND, BAT_SEND, TEST_BATON);

   #ifdef VISUAL
      Sleep(20);           // 20 milliseconds
   #else
      usleep(20000);       // 20000 microseconds = 20 milliseconds
   #endif

   // search for OK receive message
   if (MidiInput::getCount() == 0) {  
      // nothing came back via MIDI input
      setReportStatus(reportStatus);
      return 0;
   } else {
      // something came in MIDI input, check for baton alive message
      for (int i=0; i<MidiInput::getCount(); i++) {
         if ( (*this)[i].command() == BAT_MIDI_COMMAND &&
              (*this)[i].p1() == BAT_SEND &&
              (*this)[i].p2() == 0 ) {
                  setReportStatus(reportStatus);
                  return 1;
         }
      }
   }
   setReportStatus(reportStatus);
   return 0;
}



//////////////////////////////
//
// RadioDrum::toggleError -- switches the printing error
//    status when the class received a MIDI messages it
//    doesn't understand (and probably shouldn't have
//    ignored.)
//

void RadioDrum::toggleError(void) {
   errorQ = !errorQ;
}



//////////////////////////////
//
// RadioDrum::togglePositionReporting
//

void RadioDrum::togglePositionReporting(void) {
   if (reportingQ) {
      reportingQ = 0;
      positionReportingOff();
   } else {
      reportingQ = 1;
      positionReportingOn();
   }
}



//////////////////////////////
//
// RadioDrum::writeCalibrationFile -- request the calibration
//     data from the radio drum and stores the filename to which
//     the data is to be written to when it has been received from 
//     the radio drum.
//

int RadioDrum::writeCalibrationFile(const char* aFilename) {
   pushReportingStatus();
   positionReportingOff();

   // ask for calibration data from the radio drum   
   requestCalibration();
   writingCalib = 1;

   // store the filename for writeCalibrationData() function
   // which will be called when the data is received via MIDI input
   // from the radio drum.
   if (calibFilename != NULL) {
      delete [] calibFilename;
   }
   calibFilename = new char[strlen(aFilename) + 1];
   strcpy(calibFilename, aFilename);

   return 1;
}





///////////////////////////////////////////////////////////////////////////
//
// private functions
//


//////////////////////////////
//
// RadioDrum::interpretCommand -- incoming MIDI messages stored in the
//     MidiInput incoming buffer will be processed one at a time,
//     checking for MIDI messages from a radio drum.  This function
//     will know what to do with that MIDI input message.
//

void RadioDrum::interpretCommand(MidiMessage aMessage) {
   ushort value;  // for the buff value receive commands

   if (aMessage.command() == BAT_MIDI_COMMAND) {
      switch (aMessage.p1()) {
         case BAT_STICK1_RESPONSE_X:         // stick 1 responding to poll; x
            s1ps(aMessage.time);
            s1pd(DATA_X, aMessage.p2());
            break;
         case BAT_STICK1_RESPONSE_Y:         // stick 1 responding to poll; y
            s1pd(DATA_Y, aMessage.p2());
            break;
         case BAT_STICK1_RESPONSE_Z:         // stick 1 responding to poll; z
            s1pd(DATA_Z, aMessage.p2());
            break;
         case BAT_STICK2_RESPONSE_X:         // stick 2 responding to poll; x
            s2ps(aMessage.time);
            s2pd(DATA_X, aMessage.p2());
            break;
         case BAT_STICK2_RESPONSE_Y:         // stick 2 responding to poll; y
            s2pd(DATA_Y, aMessage.p2());
            break;
         case BAT_STICK2_RESPONSE_Z:         // stick 2 responding to poll; z
            s2pd(DATA_Z, aMessage.p2());
            break;
         case BAT_POT1_RESPONSE:             // pot 1 responding to poll
            d1p = aMessage.p2();             // update global state variable
            recordState(aMessage.time, DIAL1RECORD, d1p);
            d1pb.insert(aMessage.p2());
            break;
         case BAT_POT2_RESPONSE:             // pot 2 responding to poll
            d2p = aMessage.p2();             // update global state variable
            recordState(aMessage.time, DIAL2RECORD, d2p);
            d2pb.insert(aMessage.p2());
            break;
         case BAT_POT3_RESPONSE:             // pot 3 responding to poll
            d3p = aMessage.p2();             // update global state variable
            recordState(aMessage.time, DIAL3RECORD, d3p);
            d3pb.insert(aMessage.p2());
            break;
         case BAT_POT4_RESPONSE:             // pot 4 responding to poll
            d4p = aMessage.p2();             // update global state variable
            recordState(aMessage.time, DIAL4RECORD, d4p);
            d4pb.insert(aMessage.p2());
            break;
         case BAT_STICK1_TRIGGER:            // stick 1 got triggered
            s1ts(aMessage.time);
            s1td(DATA_W, aMessage.p2());
            break;
         case BAT_STICK1_TRIG_X:             // stick 1 got triggered
            s1td(DATA_X, aMessage.p2());
            break;
         case BAT_STICK1_TRIG_Y:             // stick 1 got triggered
            s1td(DATA_Y, aMessage.p2());
            break;
         case BAT_STICK2_TRIGGER:            // stick 2 got triggered
            s2ts(aMessage.time);
            s2td(DATA_W, aMessage.p2());
            break;
         case BAT_STICK2_TRIG_X:             // stick 2 got triggered
            s2td(DATA_X, aMessage.p2());
            break;
         case BAT_STICK2_TRIG_Y:             // stick 2 got triggered
            s2td(DATA_Y, aMessage.p2());
            break;

         case BAT_BUTTON_FOOT_TRIGGER:       // button or pedal was triggered
            switch (aMessage.p2()) {
               case BAT_B14p_TRIGGER:        // B14+ button pressed
                  b14pt = aMessage.time;
                  b14ptb.insert(b14pt);
                  recordState(aMessage.time, BUTTON1RECORD);
                  b14plustrig();
                  break;
               case BAT_B15p_TRIGGER:        // B15+ button pressed
                  b15pt = aMessage.time;
                  recordState(aMessage.time, BUTTON2RECORD);
                  b15plustrig();
                  b15ptb.insert(b15pt);
                  break;
               case BAT_B14m_DOWN_TRIGGER:   // B14- pedal was depressed
                  b14mdt = aMessage.time;
                  b14mdtb.insert(b14mdt);
                  recordState(b14mdt, FOOTPEDAL1RECORD, 1);
                  b14minusdowntrig();
                  break;
               case BAT_B14m_UP_TRIGGER:     // B14- pedal was released
                  b14mut = aMessage.time;
                  b14mutb.insert(b14mut);
                  recordState(b14mut, FOOTPEDAL1RECORD, 0);
                  b14minusuptrig();
                  break;
               case BAT_B15m_DOWN_TRIGGER:   // B15- pedal was depressed
                  b15mdt = aMessage.time;
                  b15mdtb.insert(b15mdt);
                  recordState(b15mut, FOOTPEDAL2RECORD, 1);
                  b15minusdowntrig();
                  break;
               case BAT_B15m_UP_TRIGGER:     // B15- pedal was released
                  b15mut = aMessage.time;
                  b15mutb.insert(b15mut);
                  recordState(b15mut, FOOTPEDAL2RECORD, 0);
                  b15minusuptrig();
                  break;
            }
            break;

         case BAT_SEND:                     // baton alive message et al.
            // ignore this command it is handled in the testBaton function
            break;   
         default:
            if (errorQ) {
               cerr << "Baton command not recognized: 0x" << hex
                    << (int)aMessage.command() << dec
                    << ", param1 = " << (int)aMessage.p1()
                    << ", param2 = " << (int)aMessage.p2()
                    << endl;
            }
      }  // end of switch (aMessage.command())

   }  // end of if a baton command (0xa0)

   // take care of calibration data
   // nibbles are replaced as they come in from the radio drum
   else if (aMessage.command() == 0xa1) {
      // if p1 is not in the range from 0 to 21 then ignore the error:
      if (aMessage.p1() > 21) return;
      value = aMessage.p2();
      value &= 0x000f;                         // guarentee no high bits
      calib[aMessage.p1()] &= 0x0fff;          // blank out old nibble
      calib[aMessage.p1()] |= value << 12;
   } else if (aMessage.command() == 0xa2) {
      // if p1 is not in the range from 0 to 21 then ignore the error:
      if (aMessage.p1() > 21) return;
      value = aMessage.p2();
      value &= 0x000f;                         // guarentee no high bits
      calib[aMessage.p1()] &= 0xf0ff;          // blank out old nibble
      calib[aMessage.p1()] |= value << 8;
   } else if (aMessage.command() == 0xa3) {
      // if p1 is not in the range from 0 to 21 then ignore the error:
      if (aMessage.p1() > 21) return;
      value = aMessage.p2();
      value &= 0x000f;                         // guarentee no high bits
      calib[aMessage.p1()] &= 0xff0f;          // blank out old nibble
      calib[aMessage.p1()] |= value << 4;
   } else if (aMessage.command() == 0xa4) {
      // if p1 is not in the range from 0 to 21 then ignore the error:
      if (aMessage.p1() > 21) return;
      value = aMessage.p2();
      value &= 0x000f;                         // guarentee no high bits
      calib[aMessage.p1()] &= 0xfff0;          // blank out old nibble
      calib[aMessage.p1()] |= value;
 
      // write the calib file if last nibble of last element received:
      if (aMessage.p1() == 21 && writingCalib) {
         writingCalib = 0;

         // finished receiving calib data from radio drum.  The data
         // is sent only when a user request to write data to a file.
  	      // so write the file now.
         if (checkCalibration() != 0) {
            writeCalibrationData();
            cerr << "Calibration data written to file: "
                 << calibFilename << endl;
         } else {
            cerr << "Corrupt calibration.  Did not write to file." << endl;
         }
         // set the position reporting to what it was before the
         // calibration was requested from the radio drum
         popReportingStatus();
      }
   }

   // take care of buf data
   else if (aMessage.command() == 0xa5) {
      // if p1 is not in the range from 0 to 15 then ignore the error:
      if (aMessage.p1() > 15) return;
      completeBufQ[aMessage.p1()] = 0;
      value = (ushort)(aMessage.p2() & 0x000f);
      buf[aMessage.p1()] &= 0x0fff;
      buf[aMessage.p1()] |= (value << 12);
   } else if (aMessage.command() == 0xa6) {
      // if p1 is not in the range from 0 to 15 then ignore the error:
      if (aMessage.p1() > 15) return;
      value = (ushort)(aMessage.p2() & 0x000f);
      buf[aMessage.p1()] &= 0xf0ff;
      buf[aMessage.p1()] |= (value << 8);
   } else if (aMessage.command() == 0xa7) {
      // if p1 is not in the range from 0 to 15 then ignore the error:
      if (aMessage.p1() > 15) return;
      value = (ushort)(aMessage.p2() & 0x000f);
      buf[aMessage.p1()] &= 0xff0f;
      buf[aMessage.p1()] |= (value << 4);
   } else if (aMessage.command() == 0xa8) {
      // if p1 is not in the range from 0 to 15 then ignore the error:
      if (aMessage.p1() > 15) return;
      value = (ushort)(aMessage.p2() & 0x000f);
      buf[aMessage.p1()] &= 0xfff0;
      buf[aMessage.p1()] |= value;
      completeBufQ[aMessage.p1()] = 1;

      // record buffer element to file if recording
      // NOTE: have to figure out what to put in the time slots below; 0 for now.
      switch (aMessage.p1()) {
         case 0:   recordState(timer.getTime(), ANTENNA0RECORD, buf[0]);   break;
         case 1:   recordState(timer.getTime(), ANTENNA1RECORD, buf[1]);   break;
         case 2:   recordState(timer.getTime(), ANTENNA2RECORD, buf[2]);   break;
         case 3:   recordState(timer.getTime(), ANTENNA3RECORD, buf[3]);   break;
         case 4:   recordState(timer.getTime(), ANTENNA4RECORD, buf[4]);   break;
         case 5:   recordState(timer.getTime(), POT4RECORD,     buf[5]);   break;
         case 6:   recordState(timer.getTime(), ANTENNA5RECORD, buf[6]);   break;
         case 7:   recordState(timer.getTime(), ANTENNA6RECORD, buf[7]);   break;
         case 8:   recordState(timer.getTime(), ANTENNA7RECORD, buf[8]);   break;
         case 9:   recordState(timer.getTime(), ANTENNA8RECORD, buf[9]);   break;
         case 10:  recordState(timer.getTime(), ANTENNA9RECORD, buf[10]);  break;
         case 11:  recordState(timer.getTime(), POT1RECORD,     buf[11]);  break;
         case 12:  recordState(timer.getTime(), POT2RECORD,     buf[12]);  break;
         case 13:  recordState(timer.getTime(), POT3RECORD,     buf[13]);  break;
         case 14:  recordState(timer.getTime(), B14RECORD,      buf[14]);  break;
         case 15:  recordState(timer.getTime(), B15RECORD,      buf[15]);  break;
      }
   }

   else if (aMessage.command() == 0xaa) {              // miditest
      currentmidi = aMessage.p2();
      if (currentmidi != 0) {
         if (currentmidi != lastmidi + 1) {
            midierrors++;
         }
         lastmidi = currentmidi;
         miditests++;
      }
   } else if (aMessage.command() = 0xab) {             // miditest
      currentmidi = aMessage.p2();
      if (currentmidi != 0) {
         if (currentmidi != lastmidi+1) { 
            midierrors++;
         }
         lastmidi = currentmidi;
         miditests++;
      }
   }


   // All other MIDI commands are ignored.  Maybe they are
   // for/from a synthesizer or something like that.

}



//////////////////////////////
//
// RadioDrum::pushReportingStatus
//

void RadioDrum::pushReportingStatus(void) {
   storedReportStatus = reportingQ;
}



//////////////////////////////
//
// RadioDrum::popReportingStatus
//
 
void RadioDrum::popReportingStatus(void) {
   if (storedReportStatus == reportingQ)  {
      return;
   } else { 
      togglePositionReporting();
   }
}



//////////////////////////////
//
// RadioDrum::recordState
//

void RadioDrum::recordState(long aTime, const char* aState) {
   if (recordingQ()) {
      if (recordTimeOffset == -1) {
         recordTimeOffset = aTime;
      }
      recordOutput << aTime-recordTimeOffset 
                   << '\t' << aState << '\n';
   }
}


void RadioDrum::recordState(long aTime, const char* aState, int value) {
   if (recordingQ()) {
      if (recordTimeOffset == -1) {
         recordTimeOffset = aTime;
      }
      recordOutput << aTime - recordTimeOffset << '\t' 
                   << aState << '\t' << value << '\n';
   }
}


void RadioDrum::recordState(long aTime, const char* aState, 
      int value1, int value2, int value3) {
   if (recordingQ()) {
      if (recordTimeOffset == -1) {
         recordTimeOffset = aTime;
      }
      recordOutput << aTime - recordTimeOffset << '\t' << aState 
                   << '\t' << value1 << '\t' << value2 
                   << '\t' << value3 << '\n';
   }
}



//////////////////////////////
//
// RadioDrum::sendCalibration -- Sends the 
//    calibration data stored in object to the
//    radio drum.  size of calib array is 22.
//

int RadioDrum::sendCalibration(void) {
   int status = checkCalibration();
   if (status != 0) {
      for (int j=0; j<22; j++) {
         MidiOutput::rawsend(0xa1, j, (calib[j] >> 12) & 0x000f);
         MidiOutput::rawsend(0xa2, j, (calib[j] >> 8)  & 0x000f);
         MidiOutput::rawsend(0xa3, j, (calib[j] >> 4)  & 0x000f);
         MidiOutput::rawsend(0xa4, j, (calib[j])       & 0x000f);
      }
   }
   return status;
}



//////////////////////////////
//
// RadioDrum::s1td -- "Stick 1 Trigger Data"
//    Stores trigger position data in temporary storage
//    until a complete set of tigger data is received.
//

void RadioDrum::s1td(int flag, uchar aValue) {
   switch (flag) {
      case DATA_W:
         w1t = aValue;         // set the immediate state variable
                               // whack1, trigtime1 set in the function s1ts
         s1tf += 1;            // set flag bit 1 (or create an error)
         break;
      case DATA_X:
         x1t = aValue;         // set the immediate state variable
         whack1x = 1;          // set the global state variable
         s1tf += 4;            // set flag bit 3 (or create an error)
         break;
      case DATA_Y:
         y1t = aValue;         // set the immediate state variable
         whack1y = 1;          // set the global state variable
         s1tf += 16;           // set flag bit 5 (or create an error)
         if (s1tf == 21) {     // if complete data then store trigger set
            t1tb.insert((uchar)t1t);
            w1tb.insert((uchar)w1t);
            x1tb.insert((uchar)x1t);
            y1tb.insert((uchar)y1t);
            stick1trig();               // call the user-defined behavior function
            recordState(t1t, TRIGGER1RECORD, x1t, y1t, w1t);
         } else {
//            cerr << "Error: incomplete stick 1 position information" 
//                 << ", flag value is: " << (int)s1tf << endl;
         }
         break;
   }
}



//////////////////////////////
//
// RadioDrum::s1ts -- "Stick 1 Trigger Status"
//    Keeps track of a compete set of stick 1 trigger positions.  
//    This function resets temporary values to a new set of 
//    stick 1 trigger values.   Stores the timestamp and initilaizes
//    the s1tf status flag.
//

void RadioDrum::s1ts(long aTime) {
   trigger1 = 1;               // set the global state variable
   t1t = aTime;                // set the internal temporary value
   s1tf = 0x00;                // initialize the flag to determine good set
}



//////////////////////////////
//
// RadioDrum::s2td -- "Stick 2 Trigger Data"
//    Stores trigger position data in temporary storage
//    until a complete set of tigger data is received.
//

void RadioDrum::s2td(int flag, uchar aValue) {
   switch (flag) {
      case DATA_W:
         w2t = aValue;          // set the global state variable
                                // whack2, trigtime2 set in s2ts function
         s2tf += 1;             // set flag bit 1 (or create an error)
         break;
      case DATA_X:
         x2t = aValue;          // set the global state variable
         whack2x = 1;           // set the global state variable
         s2tf += 4;             // set flag bit 3 (or create an error)
         break;
      case DATA_Y:
         y2t = aValue;          // set the global state variable
         whack2y = 1;           // set the global state variable
         s2tf += 16;            // set flag bit 5 (or create an error)
         if (s2tf == 21) {      // if complete trigger, then store
            t2tb.insert(t2t); 
            w2tb.insert((uchar)w2t);
            x2tb.insert((uchar)x2t);
            y2tb.insert((uchar)y2t);
            stick2trig();               // call the user state function
            recordState(t2t, TRIGGER2RECORD, x2t, y2t, w2t);
         } else {
//            cerr << "Error: incomplete stick 2 trigger information" 
//                 << ", flag value is: " << (int)s2tf << endl;
         }
         break;
   }
}



//////////////////////////////
//
// RadioDrum::s2ts -- "Stick 2 Trigger Status"
//    Keeps track of a compete set of stick 2 trigger positions.  
//    This function reset temporary values to new set of 
//    stick 2 trigger values.   Stores the timestamp and initilaizes
//    the s1tf status flag.
//

void RadioDrum::s2ts(long aTime) {
   trigger2 = 1;               // set the global state variable
   t2t = aTime;           // store the internal temporary value
   s2tf = 0x00;                // initialize the flag to determine good set
}



//////////////////////////////
//
// RadioDrum::s1pd -- "Stick 1 Position Data"
//    Stores position data in temporary storage
//    until a complete set of position data is received.
//

void RadioDrum::s1pd(int flag, uchar aValue) {
   switch (flag) {
      case DATA_X:
         x1p = aValue;         // set the global state variable
         s1pf += 1;            // set bit 1 (or create an error)
         break;
      case DATA_Y:
         y1p = aValue;          // set the global state variable
         s1pf += 4;             // set bit 3 (or create an error)
         break;
      case DATA_Z:
         z1p = aValue;          // set the global state variable
         s1pf += 16;            // set bit 5 (or create an error)
         if (s1pf == 21) {      // if complete temp position set then store
            x1pb.insert((uchar)x1p);
            y1pb.insert((uchar)y1p);
            z1pb.insert((uchar)z1p);
			stick1position();
            recordState(t1p, POSITION1RECORD, x1p, y1p, z1p);
         } else {
//            cerr << "Error: incomplete stick 1 position information"  
//                 << ", flag value is: " << (int)s1pf << endl;
         }
         break;
   }
}



//////////////////////////////
//
// RadioDrum::s1ps -- "Stick 1 Position Status"
//    Keeps track of a compete set of stick 1 positions.  
//    This function reset temporary values to new set of 
//    stick 1 position values.   Stores the timestamp and initilaizes
//    the s1pf status flag.
//

void RadioDrum::s1ps(long aTime) {
   t1p = aTime;
   t1pb.insert(t1p); 
   s1pf = 0x00;
}




//////////////////////////////
//
// RadioDrum::s2pd -- "Stick 2 Position Data"
//    Stores position data in temporary storage
//    until a complete set of position data is received.
//

void RadioDrum::s2pd(int flag, uchar aValue) {
   switch (flag) {
      case DATA_X:
         x2p = aValue;          // set the global state variable
         s2pf += 1;             // set flag bit 1 (or create an error)
         break;
      case DATA_Y:
         y2p = aValue;          // set the global state variable
         s2pf += 4;             // set flag bit 3 (or create an error)
         break;
      case DATA_Z:
         z2p = aValue;          // set the global state variable
         s2pf += 16;            // set flag bit 5 (or create an error)
         if (s2pf == 21) {      // if complete temp position set then store
            x2pb.insert((uchar)x2p);
            y2pb.insert((uchar)y2p);
            z2pb.insert((uchar)z2p);
			stick2position();
            recordState(t2p, POSITION2RECORD, x2p, y2p, z2p);
         } else {
//            cerr << "Error: incomplete stick 2 position information" 
//                 << ", flag value is: " << (int)s2pf << endl;          
         }
         break;
   }
}



//////////////////////////////
//
// RadioDrum::s2ps -- "Stick 2 Position Status"
//    Keeps track of a compete set of stick 2 positions.  
//    This function reset temporary values to new set of 
//    stick 2 position values.   Stores the timestamp and initilaizes
//    the s1pf status flag.
//

void RadioDrum::s2ps(long aTime) {
   t2p = aTime;
   t2pb.insert(t2p);
   s2pf = 0x00;
}



//////////////////////////////
//
// RadioDrum::stoprunningstatus -- some smarty-pants MIDI drivers
//     will go into running status by themselves.  This is a no-no
//     for the radio baton (the physical device) since it does
//     not understand running status (where the MIDI command is
//     not specified if the same command came just before).
//

void RadioDrum::stoprunningstatus(void) {
   rawsend(0xaf, 0, 0);
}



//////////////////////////////
//
// RadioDrum::writeCalibrationData -- writes calibration data
//	from calib array.  This function is called after
//	a successful transmission of calib data array from
//	the radio drum  or you can just write the current 
//	contents of the array as it is stored in the computer
//	(which may be different from the current state in the
//	radio drum.
//

void RadioDrum::writeCalibrationData(void) {
   // successful receipt of calibration, so write it to a file:
   fstream output(calibFilename, ios::out);
   if (!output.is_open()) {
      cerr << "Error: could not open " << calibFilename 
           << " to write calibration data." << endl;
      // not a fatal error, however
      return;
   }

   output << "hit1      " << calib[0] << endl;
   output << "set1      " << calib[1] << endl;
   output << "hit2      " << calib[2] << endl;
   output << "set2      " << calib[3] << endl;
   output << "zero1x    " << calib[4] << endl;
   output << "zero1y    " << calib[5] << endl;
   output << "zero2x    " << calib[6] << endl;
   output << "zero2y    " << calib[7] << endl;
   output << "range     " << calib[8] << endl;
   output << "restpoint " << calib[9] << endl;
   for (int i=10; i<21; i++) {   // 22 elements in calib array
       output << calib[i] << endl;
   }
   output << "zcalib    " << calib[21] << endl;

}



//////////////////////////////
//
// RadioBatonEmptyBehavior -- this function is the default function
//     to which the radio baton class behavior functions point to.
// 

void RadioBatonEmptyBehavior(void) {
   // nothing to do since it is empty
}
